//ModbusRTU.cpp

#include <Arduino.h>
#include "config.h"
#include "ModbusRTU.h"
#include "maindat.h"

DAT_MRTU  dmrtu;
//uint8_t wcnt;	//количество прочитанных байтов

//инициализация
int8_t  sMRTU_INIT(void)
{
  memset(&dmrtu.rxbuf[0], 0, _szbuf);
  memset(&dmrtu.txbuf[0], 0, _szbuf);
  dmrtu.adrx = 0;
  dmrtu.adtx = 0;
  return 0;
}

//перезапуск последовательного порта
void RestartSerial(void)
{
	Serial.end();
	delay(_sMRTU_Timeout_4);
	Serial.begin(_sMRTU_Baudrate);
	Serial.setTimeout(_sMRTU_Timeout_4);
}

/***************************/

//контрольная сумма
inline uint16_t sMRTU_CRC(uint8_t* data, uint16_t len)
{
#if(_CRC_USE_TABLES == 0)
	uint16_t res_CRC = 0xFFFF;
	uint16_t count = 0;
	uint8_t count_crc;
	uint8_t dt;
	while(count < len)
	{
		//
		count_crc = 0;
		dt = (uint8_t)(data[count]);
		res_CRC ^= (uint16_t)(dt);
		//
		while(count_crc < 8)
		{
			if((res_CRC & 0x0001) < 1)
			{
				res_CRC = (res_CRC >> 1) & 0x7FFF;
			}
			else
			{
				res_CRC = (res_CRC >> 1) & 0x7FFF;
				res_CRC ^= 0xA001;
			};
			count_crc++;
		};
		count++;
	}
  	return (res_CRC);
#elif(_CRC_USE_TABLES == 1)
	uint8_t uchCRCHi = 0xFF;
	uint8_t uchCRCLo = 0xFF;
	uint16_t uIndex ;
	while(len--)
	{
		uIndex = uchCRCHi ^ *data++ ;
		uchCRCHi = uchCRCLo ^ auchCRCHi[uIndex] ;
		uchCRCLo = auchCRCLo[uIndex] ;
	}
	//return (uint16_t)(uchCRCHi | uchCRCLo << 8) ;
	return (uint16_t)(uchCRCHi | uchCRCLo * 256) ;
#endif
}

//контрольная сумма для ответа
__inline void sMRTU_CRCo(uint8_t* data, uint16_t len, uint8_t * crch, uint8_t * crcl)
{
	uint8_t uchCRCHi = 0xFF;
	uint8_t uchCRCLo = 0xFF;
	uint16_t uIndex ;
	while(len--)
	{
		uIndex = uchCRCHi ^ *data++ ;
		uchCRCHi = uchCRCLo ^ auchCRCHi[uIndex] ;
		uchCRCLo = auchCRCLo[uIndex] ;
	}
	*crch = uchCRCLo;
	*crcl = uchCRCHi;
}

//контрольная сумма для приёма - результат сравнение: 0 - совпало
__inline uint8_t sMRTU_CRCi(uint8_t * data, uint16_t len, uint8_t * crch, uint8_t * crcl)
{
	uint8_t uchCRCHi = 0xFF;
	uint8_t uchCRCLo = 0xFF;
	uint16_t uIndex ;
	//uint8_t res;
	while(len--)
	{
		uIndex = uchCRCHi ^ *data++ ;
		uchCRCHi = uchCRCLo ^ auchCRCHi[uIndex] ;
		uchCRCLo = auchCRCLo[uIndex] ;
	}
	return (*crch ^ uchCRCLo) | (*crcl ^ uchCRCHi);
}

/*
****************************
CMD3 - чтение двоичного
содержания регистров
****************************
*/
inline void	MRTU_CMD3(void)
{
	uint8_t	  	cntbytes;					//счётчик байт в поле данных
	int		addr;
	uint8_t	  	bsreg	= dmrtu.rxbuf[_ADDR_03_START_REG_L];
	uint8_t   	bcntreg	= dmrtu.rxbuf[_ADDR_03_CNT_REG_L];
	//проверка контрольной суммы
	if(sMRTU_CRCi(&dmrtu.rxbuf[0], 6, &dmrtu.rxbuf[_ADDR_03_INCRC_H], &dmrtu.rxbuf[_ADDR_03_INCRC_L]) == 0)
      	{
        	/*
        	****************************
        	проверка границ данных
        	****************************
        	*/
		//проверка предельного размера
		if(bcntreg >= _cntreg)
		{
			dmrtu.txbuf[0] = _MRTU_Addr;
			dmrtu.txbuf[1] = 3 | _CMD_ERR_MODE;
			dmrtu.txbuf[2] = _ILLEGAL_DATA_ADDRESS;
			sMRTU_CRCo(&dmrtu.txbuf[0], 3, &dmrtu.txbuf[3], &dmrtu.txbuf[4]);
			Serial.write(&dmrtu.txbuf[0], 5);
			return;
		}
		//movectrl
		else if((bsreg == _MoveControl) && 
			(bcntreg == _MoveControlSZ))
		{
			dmrtu.txbuf[_ADDR_MADDR]  		=  _MRTU_Addr;		//адрес устройства
			dmrtu.txbuf[_ADDR_MCMD]  		=  3;			//команда
			dmrtu.txbuf[_ADDR_03_CNT_BYTES]		= 16;			//счётчик байтов для передачи
			addr					= (int)(&rdmovectrl);
			cntbytes				= _ADDR_03_DATA;		//счётчик байтов + смещение
			while(cntbytes <  (_ADDR_03_DATA + 16))
			{
				dmrtu.txbuf[cntbytes++]	= (*(uint8_t*)(addr + 1));
				dmrtu.txbuf[cntbytes++]	= (*(uint8_t*)(addr));
				addr += 2;
			}
			sMRTU_CRCo(&dmrtu.txbuf[0], 19, &dmrtu.txbuf[19], &dmrtu.txbuf[19 + 1]);
			Serial.write(&dmrtu.txbuf[0], 19 + 2);
			return;
		}
		//ComMotion
		else if((bsreg >= _ComMotion_RMS) && 
			(bsreg < (_ComMotion_RMS + _ComMotion_SzRM)) && 
			((bsreg + bcntreg) <= (_ComMotion_RMS + _ComMotion_SzRM)))
		{
			dmrtu.txbuf[_ADDR_MADDR]  	= _MRTU_Addr;			//адрес устройства
			dmrtu.txbuf[_ADDR_MCMD]  	= 3;				//команда
			dmrtu.txbuf[_ADDR_03_CNT_BYTES]	= bcntreg << 1;			//счётчик байтов для передачи
			//
			bsreg 				-= _ComMotion_RMS;		//стартовый регистр
			cntbytes			= _ADDR_03_DATA;		//счётчик байтов + смещение
			bcntreg				+= bsreg;			//количество регистров со смещением
			addr				= (int)(&dmotor[bsreg]);
			while(bsreg < bcntreg)
			{	
				dmrtu.txbuf[cntbytes++] = (*(uint8_t*)(addr + 1));
				dmrtu.txbuf[cntbytes++] = (*(uint8_t*)(addr));
				bsreg++;
				addr += 2;
			}
			sMRTU_CRCo(&dmrtu.txbuf[0], cntbytes, &dmrtu.txbuf[cntbytes], &dmrtu.txbuf[cntbytes + 1]);
			Serial.write(&dmrtu.txbuf[0], cntbytes + 2);
			return;
		}
		//Multiservo
		else if((bsreg >= _Multiservo_RSS1) && 
			(bsreg < (_Multiservo_RSS1 + _Multiservo_SzRS)) && 
			((bsreg + bcntreg) <= (_Multiservo_RSS1 + _Multiservo_SzRS)))
		{
			dmrtu.txbuf[_ADDR_MADDR]  	=  _MRTU_Addr;			//адрес устройства
			dmrtu.txbuf[_ADDR_MCMD]  	=  3;				//команда
			dmrtu.txbuf[_ADDR_03_CNT_BYTES]	= bcntreg << 1;			//счётчик байтов для передачи
			//
			bsreg 				-= _Multiservo_RSS1;		//стартовый регистр
			cntbytes			= _ADDR_03_DATA;		//счётчик байтов + смещение
			bcntreg				+= bsreg;			//количество регистров со смещением
			addr				= (int)(&dservo[bsreg]);
			while(bsreg < bcntreg)
			{
				dmrtu.txbuf[cntbytes++] = (*(uint8_t*)(addr + 1));
				dmrtu.txbuf[cntbytes++] = (*(uint8_t*)(addr));
				bsreg++;
				addr += 2;
			}
			sMRTU_CRCo(&dmrtu.txbuf[0], cntbytes, &dmrtu.txbuf[cntbytes], &dmrtu.txbuf[cntbytes + 1]);
			Serial.write(&dmrtu.txbuf[0], cntbytes + 2);
			return;
		}
		//Измерение времени цикла
		else if((bsreg == _timeCycleMax) && (bcntreg == 1))
		{
			dmrtu.txbuf[_ADDR_MADDR]  	= _MRTU_Addr;			//адрес устройства
			dmrtu.txbuf[_ADDR_MCMD]  	= 3;				//команда
			dmrtu.txbuf[_ADDR_03_CNT_BYTES]	= 2;				//счётчик байтов для передачи
			dmrtu.txbuf[_ADDR_03_DATA]	= (timeCycleMax >> 8);
			dmrtu.txbuf[_ADDR_03_DATA + 1]	= (timeCycleMax & 0xFF);
			sMRTU_CRCo(&dmrtu.txbuf[0], 5, &dmrtu.txbuf[5], &dmrtu.txbuf[6]);
			Serial.write(&dmrtu.txbuf[0], 7);
			return;
		}
		else
		{
			dmrtu.txbuf[0] = _MRTU_Addr;
			dmrtu.txbuf[1] = 3 | _CMD_ERR_MODE;
			dmrtu.txbuf[2] = _ILLEGAL_DATA_ADDRESS;
			sMRTU_CRCo(&dmrtu.txbuf[0], 3, &dmrtu.txbuf[3], &dmrtu.txbuf[4]);
			Serial.write(&dmrtu.txbuf[0], 5);
			return;
		}
	}
}

/*
****************************
CMD6 - запись в один регистр
****************************
*/
inline void	MRTU_CMD6(void)
{
	uint16_t  sreg;			//стартовый регистр
	uint8_t	  bsreg;
	uint16_t  datreg;
      	//стартовый регистр
	sreg	= (uint16_t)(dmrtu.rxbuf[_ADDR_06_ADDR_REG_H] * 256)
      		| (uint16_t)(dmrtu.rxbuf[_ADDR_06_ADDR_REG_L]);
      	//данные регистра
	datreg  = (uint16_t)(dmrtu.rxbuf[_ADDR_06_DAT_H] * 256)
      		| (uint16_t)(dmrtu.rxbuf[_ADDR_06_DAT_L]);
	//проверка контрольной суммы
	if(sMRTU_CRCi(&dmrtu.rxbuf[0], 6, &dmrtu.rxbuf[_ADDR_06_INCRC_H], &dmrtu.rxbuf[_ADDR_06_INCRC_L]) == 0)
      	{
        	/*
        	****************************
        	проверка границ данных
        	****************************
        	*/
		bsreg = (uint8_t)(sreg);
		//контрольный регистр
		if(bsreg == _CntrlReg)
		{
			CntrlReg	= (int16_t)(datreg);
			memcpy(&dmrtu.txbuf[0], &dmrtu.rxbuf[0], 8);
           		Serial.write(&dmrtu.txbuf[0], 8);
			return;
		}
        	//ComMotion
        	else if((bsreg >= _ComMotion_RMS) && (bsreg < (_ComMotion_RMS + _ComMotion_SzRM)))
        	{
           		dmotor[sreg - _ComMotion_RMS] = (int16_t)(datreg);
           		memcpy(&dmrtu.txbuf[0], &dmrtu.rxbuf[0], 8);
           		Serial.write(&dmrtu.txbuf[0], 8);
			return;
        	}
		//Multiservo
		else if((bsreg >= _Multiservo_RSS1) && (bsreg < (_Multiservo_RSS1 + _Multiservo_SzRS)))
		{
			dservo[sreg - _Multiservo_RSS1] = (int16_t)(datreg);
			memcpy(&dmrtu.txbuf[0], &dmrtu.rxbuf[0], 8);
			Serial.write(&dmrtu.txbuf[0], 8);
			return;
		}
		// ILLEGAL DATA ADDRESS - вышли за предел данных
		else
		{
			dmrtu.txbuf[0] = _MRTU_Addr;
			dmrtu.txbuf[1] = 6 | _CMD_ERR_MODE;
			dmrtu.txbuf[2] = _ILLEGAL_DATA_ADDRESS;
			sMRTU_CRCo(&dmrtu.txbuf[0], 3, &dmrtu.txbuf[3], &dmrtu.txbuf[4]);
			Serial.write(&dmrtu.txbuf[0], 5);
			return;
		}
        //
	}
}

/*
****************************
CMD16 - запись в несколько 
регистров
****************************
*/
inline void	MRTU_CMD16(void)
{
	int		addr;
	uint8_t   	wcntb		= 0;
	uint8_t	  	bsreg		= dmrtu.rxbuf[_ADDR_16_ADDR_REG_L];
	uint8_t   	bcntreg		= dmrtu.rxbuf[_ADDR_16_CNT_REG_L];
	uint8_t	  	cntbytes 	= dmrtu.rxbuf[_ADDR_16_CNT_BYTES];
	//проверка контрольной суммы
	if(sMRTU_CRCi(&dmrtu.rxbuf[0], cntbytes + 7, &dmrtu.rxbuf[cntbytes + 7], &dmrtu.rxbuf[cntbytes + 8]) == 0)
	{
		/*
        	****************************
        	проверка границ данных
        	****************************
        	*/
		//проверка предельного размера
		if(bcntreg >= _cntreg)
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16 | _CMD_ERR_MODE;
			dmrtu.txbuf[2]  =  _ILLEGAL_DATA_ADDRESS;
			sMRTU_CRCo(&dmrtu.txbuf[0], 3, &dmrtu.txbuf[3], &dmrtu.txbuf[4]);
			Serial.write(&dmrtu.txbuf[0], 5);
			return;
		}
		//movectrl
		else if((bsreg == _MoveControl) && 
			(bcntreg == _MoveControlSZ))
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16;
			dmrtu.txbuf[2]  =  0;			//стартовый регистр
			dmrtu.txbuf[3]  =  _MoveControl;	//стартовый регистр
			dmrtu.txbuf[4]  =  0;
			dmrtu.txbuf[5]	=  _MoveControlSZ;	//количество регистров
			addr		= (int)(&movectrl);
			wcntb		= _ADDR_16_DAT_H;
			while(wcntb < (_ADDR_16_DAT_H + 16))
			{
				(*(uint8_t*)(addr++))	=  dmrtu.rxbuf[wcntb + 1];
				(*(uint8_t*)(addr++))	=  dmrtu.rxbuf[wcntb];
				wcntb += 2;
			}
			sMRTU_CRCo(&dmrtu.txbuf[0], 6, &dmrtu.txbuf[6], &dmrtu.txbuf[7]);
			Serial.write(&dmrtu.txbuf[0], 8);
			flag_movectrl	= 0xFF;
			return;
		}
        	//ComMotion
		else if((bsreg >= _ComMotion_RMS) && 
			(bsreg < (_ComMotion_RMS + _ComMotion_SzRM)) && 
			((bsreg + bcntreg) <= (_ComMotion_RMS + _ComMotion_SzRM)))
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16;
			dmrtu.txbuf[2]  =  0;
			dmrtu.txbuf[3]  =  bsreg;
			dmrtu.txbuf[4]  =  0;
			dmrtu.txbuf[5]  =  bcntreg;
			bsreg 	-= _ComMotion_RMS;	//стартовый регистр
			wcntb 	+= _ADDR_16_DAT_H;	//счётчик байтов + смещение
			bcntreg	+= bsreg;		//количество регистров со смещением
			addr	= (int)&dmotor[bsreg];
			while(bsreg < bcntreg)
			{
				//dmotor[bsreg++]	= ((int16_t)(dmrtu.rxbuf[wcntb]) * 256)
				//		| (int16_t)(dmrtu.rxbuf[++wcntb]);
				(*(uint8_t*)(addr++))	= dmrtu.rxbuf[wcntb + 1];
				(*(uint8_t*)(addr++))	= dmrtu.rxbuf[wcntb];
				bsreg++;
				wcntb	+= 2; 
			}
			sMRTU_CRCo(&dmrtu.txbuf[0], 6, &dmrtu.txbuf[6], &dmrtu.txbuf[7]);
			Serial.write(&dmrtu.txbuf[0], 8);
			return;
		}
		//Multiservo
		else if((bsreg >= _Multiservo_RSS1) && 
			(bsreg < (_Multiservo_RSS1 + _Multiservo_SzRS)) && 
			((bsreg + bcntreg) <= (_Multiservo_RSS1 + _Multiservo_SzRS)))
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16;
			dmrtu.txbuf[2]  =  0;			//стартовый регистр
			dmrtu.txbuf[3]  =  bsreg;		//стартовый регистр
			bsreg 		-= _Multiservo_RSS1;	//стартовый регистр
			wcntb 		+= _ADDR_16_DAT_H;	//счётчик байтов + смещение
			dmrtu.txbuf[4]  =  0;
			dmrtu.txbuf[5]	=  bcntreg;		//количество регистров
			bcntreg		+= bsreg;		//количество регистров со смещением
			addr		= (int)&dservo[bsreg];
			while(bsreg < bcntreg)
			{
				(*(uint8_t*)(addr++))	= dmrtu.rxbuf[wcntb + 1];
				(*(uint8_t*)(addr++))	= dmrtu.rxbuf[wcntb];
				bsreg++;
				wcntb += 2;
			}
			sMRTU_CRCo(&dmrtu.txbuf[0], 6, &dmrtu.txbuf[6], &dmrtu.txbuf[7]);
			Serial.write(&dmrtu.txbuf[0], 8);
			return;
		}
		// ILLEGAL DATA ADDRESS - вышли за предел данных
		else
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16 | _CMD_ERR_MODE;
			dmrtu.txbuf[2]  =  _ILLEGAL_DATA_ADDRESS;
			sMRTU_CRCo(&dmrtu.txbuf[0], 3, &dmrtu.txbuf[3], &dmrtu.txbuf[4]);
			Serial.write(&dmrtu.txbuf[0], 5);
			return;
		}
	}
}

/*
****************************
неопределённая команда
****************************
*/
inline void	MRTU_CMDunknown(void)
{
	//uint16_t  crc; //подсчитанная контрольная сумма
	dmrtu.txbuf[0]  =  _MRTU_Addr;
       	dmrtu.txbuf[1]  =  dmrtu.rxbuf[1] | _CMD_ERR_MODE;
       	dmrtu.txbuf[2]	=  _ILLEGAL_FUNCTION;
	sMRTU_CRCo(&dmrtu.txbuf[0], 3, &dmrtu.txbuf[3], &dmrtu.txbuf[4]);
       	Serial.write(&dmrtu.txbuf[0], 5);
}

/*
****************************
приём команды и обработка
****************************
*/
void  MRTU_rWORK(void)
{
	if(Serial.available() < 8)
		return;
	//
	Serial.readBytes((char*)(&dmrtu.rxbuf[0]), _szbuf);
	//не наш адрес
  	if(dmrtu.rxbuf[_ADDR_MADDR] != _MRTU_Addr)
		return;
	//определение команд
	switch(dmrtu.rxbuf[_ADDR_MCMD])
	{
		case 3:
			MRTU_CMD3();
			break;
		case 6:
			MRTU_CMD6();
			break;
		case 16:
			MRTU_CMD16();
			break;
		default:
			MRTU_CMDunknown();
			break;
	}
	Serial.flush();
}


