//ModbusRTU.h

#ifndef _smodbusrtu_h_
#define _smodbusrtu_h_

#include "config.h"
#include <Arduino.h>

#ifndef  _szbuf  //если размер буфера не определён
#define  _szbuf  0x40
#endif

//допустимое количество регистров в пакете
#define _cntreg  ((_szbuf - 8) / 2)

//адрес устройства
#ifndef  _MRTU_Addr
#define  _MRTU_Addr  16  //адрес устройства по умолчанию
#endif

//определения
#define  _ADDR_MADDR  			0x00	//адрес адреса устройства
#define  _ADDR_MCMD  			0x01	//адрес команды
//cmd 0x03
#define  _ADDR_03_START_REG_H   	0x02	//адрес начального(старшего) байта первого регистра
#define  _ADDR_03_START_REG_L   	0x03
#define  _ADDR_03_CNT_REG_H  		0x04	//адрес старшего байта количества регистров
#define  _ADDR_03_CNT_REG_L		0x05
#define  _ADDR_03_INCRC_H  		0x06	//адрес контрольной суммы
#define  _ADDR_03_INCRC_L  		0x07
#define  _ADDR_03_CNT_BYTES		0x02	//количество байт в поле данных
#define	 _ADDR_03_DATA			0x03	//адрес начала данных
//cmd 0x04
#define  _ADDR_04_START_REG_H  		0x02	//адрес начального(старшего) байта первого регистра
#define  _ADDR_04_CNT_REG_H  		0x04	//адрес старшего байта количества регистров
#define  _ADDR_04_INCRC  		0x06	//адрес контрольной суммы
//cmd 0x06
#define _ADDR_06_ADDR_REG_H  		0x02	//адрес (старший байт) регистра
#define _ADDR_06_ADDR_REG_L  		0x03
#define _ADDR_06_DAT_H  		0x04	//адрес старшего байта данных
#define _ADDR_06_DAT_L  		0x05
#define _ADDR_06_INCRC_H  		0x06	//адрес контрольной суммы
#define _ADDR_06_INCRC_L		0x07
//cmd 0x10
#define _ADDR_16_ADDR_REG_H  		0x02  	//адрес (старший байт) начального регистра
#define _ADDR_16_ADDR_REG_L		0x03
#define _ADDR_16_CNT_REG_H  		0x04  	//адрес (старший байт) количество регистров
#define _ADDR_16_CNT_REG_L  		0x05
#define _ADDR_16_CNT_BYTES		0x06	//количество байт в поле данных
#define _ADDR_16_DAT_H  		0x07  	//адрес старшего байта данных первого регистра


//флаг ошибки
#define _CMD_ERR_MODE  			0x80
//возвращаемые ошибки
#define _ILLEGAL_FUNCTION		0x01
#define _ILLEGAL_DATA_ADDRESS		0x02

#define _CRC_USE_TABLES			1	//использование табличного метода посчёта контрольной суммы

#if(_CRC_USE_TABLES == 1)

/* Table of CRC values for high–order byte*/
const uint8_t auchCRCHi[] = {
0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01,
0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81,
0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01,
0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01,
0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01,
0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
0x40
};


/* Table of CRC values for low–order byte*/
const uint8_t auchCRCLo[] = {
0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06, 0x07, 0xC7, 0x05, 0xC5, 0xC4,
0x04, 0xCC, 0x0C, 0x0D, 0xCD, 0x0F, 0xCF, 0xCE, 0x0E, 0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09,
0x08, 0xC8, 0xD8, 0x18, 0x19, 0xD9, 0x1B, 0xDB, 0xDA, 0x1A, 0x1E, 0xDE, 0xDF, 0x1F, 0xDD,
0x1D, 0x1C, 0xDC, 0x14, 0xD4, 0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3,
0x11, 0xD1, 0xD0, 0x10, 0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3, 0xF2, 0x32, 0x36, 0xF6, 0xF7,
0x37, 0xF5, 0x35, 0x34, 0xF4, 0x3C, 0xFC, 0xFD, 0x3D, 0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A,
0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38, 0x28, 0xE8, 0xE9, 0x29, 0xEB, 0x2B, 0x2A, 0xEA, 0xEE,
0x2E, 0x2F, 0xEF, 0x2D, 0xED, 0xEC, 0x2C, 0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26,
0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0, 0xA0, 0x60, 0x61, 0xA1, 0x63, 0xA3, 0xA2,
0x62, 0x66, 0xA6, 0xA7, 0x67, 0xA5, 0x65, 0x64, 0xA4, 0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F,
0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB, 0x69, 0xA9, 0xA8, 0x68, 0x78, 0xB8, 0xB9, 0x79, 0xBB,
0x7B, 0x7A, 0xBA, 0xBE, 0x7E, 0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C, 0xB4, 0x74, 0x75, 0xB5,
0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71, 0x70, 0xB0, 0x50, 0x90, 0x91,
0x51, 0x93, 0x53, 0x52, 0x92, 0x96, 0x56, 0x57, 0x97, 0x55, 0x95, 0x94, 0x54, 0x9C, 0x5C,
0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E, 0x5A, 0x9A, 0x9B, 0x5B, 0x99, 0x59, 0x58, 0x98, 0x88,
0x48, 0x49, 0x89, 0x4B, 0x8B, 0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C,
0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42, 0x43, 0x83, 0x41, 0x81, 0x80,
0x40
};

#endif

typedef struct _DAT_MRTU
{
  uint8_t  rxbuf[_szbuf];
  uint8_t  txbuf[_szbuf];
  uint8_t  adrx;  //указатель на массив буфера приёма
  uint8_t  adtx;  //указатель на массив буфера передачи
} DAT_MRTU;

extern DAT_MRTU  dmrtu;

//инициализация
extern int8_t	sMRTU_INIT(void);
//перезапуск последовательного порта
extern void	RestartSerial(void);

//контрольная сумма
inline uint16_t sMRTU_CRC(uint8_t* data, uint16_t len, uint8_t * tbuff);
//контрольная сумма для ответа
inline void 	sMRTU_CRCo(uint8_t* data, uint16_t len, uint8_t * crch, uint8_t * crcl);
//контрольная сумма для приёма - результат сравнение: 0 - совпало
inline uint8_t	sMRTU_CRCi(uint8_t* data, uint16_t len, uint8_t * crch, uint8_t * crcl);
inline void	MRTU_CMD3(void);
inline void	MRTU_CMD6(void);
inline void	MRTU_CMD16(void);
inline void 	MRTU_CMDunknown(void);
//работа по приёму
extern void MRTU_rWORK(void);


#endif  //_smodbusrtu_h_


